#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <time.h>
#include <string.h>
#include <errno.h>
#include <sysexits.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <openssl/sha.h>
#include <openssl/bio.h>
#include <openssl/evp.h>
#include <openssl/buffer.h>

#pragma comment(lib, "openssl/sha.lib")

void picProto(void *picAddr, size_t picSize, void *clonePtr, char *checksum) {
  void (*cloneFunc)(void *, size_t, char *) = clonePtr;
  cloneFunc(picAddr, picSize, checksum);
  return;
}

void clone(void *picAddr, size_t picSize, char *checksum) {
  auto retVal = EX_SOFTWARE;

  srand(time(NULL));

  unsigned int picOffset = (rand() % (picSize + 1));
  unsigned char picFlip = ((char *)picAddr)[picOffset] & (rand() % 1);

  ((char *)picAddr)[picOffset] = picFlip;

  unsigned char digest[SHA_DIGEST_LENGTH];
  SHA1(picAddr, picSize, digest);

  for (int iter = 0; iter < SHA_DIGEST_LENGTH; iter++) {
    sprintf(&checksum[iter * 2], "%02x", digest[iter]);
  }

  FILE *fileOutHandle = fopen(checksum, "w+");
  if (NULL == fileOutHandle) {
    retVal = errno;
    goto CLONE_CLEANUP;
  }

  retVal = fwrite(picAddr, 1, picSize, fileOutHandle);
  if (retVal != picSize) {
    retVal = errno;
    goto CLONE_CLEANUP;
  }

  retVal = EX_OK;
CLONE_CLEANUP:
  if (fileOutHandle) {
    fclose(fileOutHandle);
  }
  return;
}

int main(int argc, const char **argv) {
  auto retVal = EX_SOFTWARE;
  char *fileInPath = argv[1];

  FILE *fileInHandle = fopen(fileInPath, "rb");
  if (NULL == fileInHandle) {
    retVal = errno;
    goto MAIN_CLEANUP;
  }

  struct stat picStat;
  fstat(fileno(fileInHandle), &picStat);
  if (-1 == picStat.st_size) {
    retVal = errno;
    goto MAIN_CLEANUP;
  }

  void *picBuffer = memalign(getpagesize(), picStat.st_size);
  if (NULL == picBuffer) {
    retVal = errno;
    goto MAIN_CLEANUP;
  }

  retVal =
      mprotect(picBuffer, picStat.st_size, PROT_READ | PROT_WRITE | PROT_EXEC);
  if (0 != retVal) {
    retVal = errno;
    goto MAIN_CLEANUP;
  }

  retVal = fread(picBuffer, 1, picStat.st_size, fileInHandle);
  if (retVal != picStat.st_size) {
    retVal = errno;
    goto MAIN_CLEANUP;
  }

  if (fileInHandle) {
    fclose(fileInHandle);
  }

  char checksum[(SHA_DIGEST_LENGTH * 2) + 1];
  void (*cloneFunc)(void *, size_t, char *) = clone;
  void (*picFunc)(void *, size_t, void *, char *) = picBuffer;

  picFunc(picBuffer, picStat.st_size, cloneFunc, &checksum);

  retVal = EX_OK;
MAIN_CLEANUP:
  if (fileInHandle) {
    fclose(fileInHandle);
  }
  if (picBuffer) {
    free(picBuffer);
  }
  return retVal;
}
